package database

import (
	"chromagies/src/config"
	"chromagies/src/profiler"
	"database/sql"
	"fmt"
	"log"
	"time"

	_ "github.com/go-sql-driver/mysql"
)

const (
	profiler_DATABASE_CONNECT string = "Database Connect"
	profiler_DATABASE_QUERY   string = "Database Query"
	profiler_DATABASE_PING    string = "Database Ping"
)

var db *sql.DB
var dbConfig config.Database
var isInit bool

func Init() {
	if isInit {
		return
	}

	dbConfig = config.GetConfig().Database

	profiler.Register(profiler_DATABASE_CONNECT, 20, "µs")
	profiler.Register(profiler_DATABASE_QUERY, 4000, "µs")
	profiler.Register(profiler_DATABASE_PING, 20, "µs")

	isInit = true

	Connect()

	var query string = `SHOW GRANTS;`
	var rows *sql.Rows = executeQuery(query)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var res string
			rows.Scan(&res)
			fmt.Println(res)
		}
	}

}

func Terminate() {
	isInit = false
	if db != nil {
		db.Close()
		db = nil
	}
}

func Connect() {
	if !isInit {
		return
	}

	if db != nil {
		db.Close()
	}

	var err error
	var dbPath string = fmt.Sprintf("%s:%s@(%s:%d)/%s?parseTime=true", dbConfig.User, dbConfig.Password, dbConfig.Host, dbConfig.Port, dbConfig.Database)
	log.Printf("Connecting to database %s at %s:%d as %s\n", dbConfig.Database, dbConfig.Host, dbConfig.Port, dbConfig.User)

	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		db, err = sql.Open("mysql", dbPath)
		if err == nil {
			break
		}
		log.Printf("Error while connecting the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}
	profiler.Add(profiler_DATABASE_CONNECT, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Fatalln("Cannot connect to database !")
	}

	Ping()
}

func Disconnect() {
	if !isInit {
		return
	}

	if db == nil {
		return
	}

	db.Close()
}

func Ping() {
	if !isInit {
		return
	}

	if db == nil {
		return
	}

	var err error
	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		err = db.Ping()
		if err == nil {
			break
		}
		log.Printf("Error while pinging the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}
	profiler.Add(profiler_DATABASE_PING, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Fatalln("Cannot ping the database !")
	}
}

func executeQuery(query string) *sql.Rows {
	if !isInit {
		return nil
	}

	if db == nil {
		return nil
	}

	Ping()

	var err error
	var rows *sql.Rows
	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		rows, err = db.Query(query)
		if err == nil {
			break
		}
		log.Printf("Error while querying the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}
	profiler.Add(profiler_DATABASE_QUERY, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Fatalln("Cannot query the database !")
	}

	return rows
}
